/*
 *  Copyright (C) 2007 Viktor Nordell <viktor.nordell@gmail.com>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Author : Viktor Nordell <viktor.nordell@gmail.com>
 */
#include <gtk/gtk.h>
#include <gdk/gdk.h>

#include "gui_top_area.h"
#include "gui_top_area_marshallers.h"

#define MAX_NICK_SIZE 40
#define MAX_ADDR_SIZE 30

#define TOP_BUTTON_SIZE 15
#define TOP_BUTTON_MARGIN 2
#define TOP_BUTTON_TOTAL_SIZE (TOP_BUTTON_SIZE + TOP_BUTTON_MARGIN)

#define TOP_IP_TEXT_MARGIN 15
#define TOP_NICK_TEXT_MARGIN 40

#define TOP_BUTTON_LEFT_OFFSET 2
#define TOP_BUTTON_RIGTH_OFFSET (TOP_BUTTON_SIZE+TOP_BUTTON_LEFT_OFFSET)

#define TOP_RIGHT_BUTTON_PUSHED "/top-right-button-pushed.png"
#define TOP_RIGHT_BUTTON_OVER "/top-right-button-over.png"
#define TOP_RIGHT_BUTTON "/top-right-button.png"

#define TOP_LEFT_BUTTON_PUSHED "/top-left-button-pushed.png"
#define TOP_LEFT_BUTTON_OVER "/top-left-button-over.png"
#define TOP_LEFT_BUTTON "/top-left-button.png"

#define GUI_TOP_AREA_GET_PRIVATE(obj) (G_TYPE_INSTANCE_GET_PRIVATE ((obj), GUI_TYPE_TOP_AREA, GuiTopAreaPrivate))

G_DEFINE_TYPE (GuiTopArea, gui_top_area, GTK_TYPE_DRAWING_AREA);

static gboolean gui_top_area_expose 			(GtkWidget *widget, GdkEventExpose *event);
static gboolean gui_top_area_configure_event	(GtkWidget *widget, GdkEventConfigure *event);
static gboolean gui_top_area_button_press 		(GtkWidget *widget, GdkEventButton *event);
static gboolean gui_top_area_button_release 	(GtkWidget *widget, GdkEventButton *event);
static gboolean gui_top_area_motion_notify 		(GtkWidget *widget, GdkEventMotion *event);
static gboolean gui_top_area_leave_notify_event (GtkWidget *widget, GdkEventCrossing *event);

static void gui_top_area_finalize 		(GuiTopArea *top_area);
static void gui_top_area_draw_base		(GtkWidget *widget, cairo_t *cr);
static void gui_top_area_draw_button	(GtkWidget *widget, cairo_t *cr, gchar *path, gboolean left);
static void gui_top_area_draw_buttons	(GtkWidget *widget, cairo_t *cr);
static void gui_top_area_draw_nick_name	(GtkWidget* widget, cairo_t *cr, gchar* nick);
static void gui_top_area_draw_ip_address(GtkWidget* widget, cairo_t *cr, gchar* addr);

typedef struct _GuiTopAreaPrivate GuiTopAreaPrivate;

struct _GuiTopAreaPrivate
{	
	gchar *nick;
	gchar *addr;

	gboolean left_button_pushed;
	gboolean left_button_over;
	gboolean right_button_over;
	gboolean right_button_pushed;
};

enum{
	PROP_ZERO
};

enum{
	SIGNAL_LEFT_BUTTON,
	SIGNAL_RIGHT_BUTTON,
	SIGNAL_LAST
};

static guint gui_top_area_signals[SIGNAL_LAST] = { 0, 0 };

static void
gui_top_area_class_init (GuiTopAreaClass *class)
{
	GObjectClass *obj_class;
	GtkWidgetClass *widget_class;

	obj_class = G_OBJECT_CLASS (class);
	obj_class->finalize = (GObjectFinalizeFunc) gui_top_area_finalize;

	widget_class = GTK_WIDGET_CLASS (class);

	gui_top_area_signals[SIGNAL_LEFT_BUTTON] = g_signal_new (
					"left-button-clicked",
					G_OBJECT_CLASS_TYPE (obj_class),
					G_SIGNAL_RUN_FIRST,
					0,NULL, NULL,
					_top_area_marshal_VOID__VOID,
					G_TYPE_NONE, 0);

	gui_top_area_signals[SIGNAL_RIGHT_BUTTON] = g_signal_new (
					"right-button-clicked",
					G_OBJECT_CLASS_TYPE (obj_class),
					G_SIGNAL_RUN_FIRST,
					0,NULL, NULL,
					_top_area_marshal_VOID__VOID,
					G_TYPE_NONE, 0);

	widget_class->expose_event = gui_top_area_expose;
	widget_class->configure_event = gui_top_area_configure_event;
	widget_class->button_press_event = gui_top_area_button_press;
	widget_class->button_release_event = gui_top_area_button_release;
	widget_class->motion_notify_event = gui_top_area_motion_notify;
	widget_class->leave_notify_event = gui_top_area_leave_notify_event;

	g_type_class_add_private (obj_class, sizeof (GuiTopAreaPrivate));
}

static void
gui_top_area_init (GuiTopArea *top_area)
{
	GuiTopAreaPrivate *priv;

	priv = GUI_TOP_AREA_GET_PRIVATE (top_area);

	priv->nick = NULL;
	priv->addr = NULL;

	priv->left_button_pushed 	= FALSE;
	priv->left_button_over		= FALSE;
	priv->right_button_over		= FALSE;
	priv->right_button_pushed	= FALSE;

	gtk_widget_set_extension_events (GTK_WIDGET (top_area), GDK_EXTENSION_EVENTS_CURSOR);

	gtk_widget_add_events (GTK_WIDGET (top_area), GDK_EXPOSURE_MASK
			 | GDK_LEAVE_NOTIFY_MASK
			 | GDK_BUTTON_PRESS_MASK
			 | GDK_BUTTON_RELEASE_MASK
			 | GDK_BUTTON_MOTION_MASK
			 | GDK_POINTER_MOTION_MASK);
}

static void
gui_top_area_finalize (GuiTopArea *top_area)
{
	GuiTopAreaPrivate *priv;

	priv = GUI_TOP_AREA_GET_PRIVATE (top_area);
	g_free(priv->nick);
	g_free(priv->addr);
}

static gboolean
gui_top_area_configure_event (GtkWidget *top_area, GdkEventConfigure *event)
{
	return FALSE;
}

static gboolean
gui_top_area_expose (GtkWidget *widget, GdkEventExpose *event)
{
	cairo_t *cr;
	GuiTopAreaPrivate *priv;

	priv = GUI_TOP_AREA_GET_PRIVATE (widget);    
    
    cr = gdk_cairo_create(widget->window);	
        
    cairo_rectangle(cr, event->area.x, event->area.y,
		    		event->area.width, event->area.height);
    cairo_clip(cr);	
    
    gui_top_area_draw_base(widget, cr);		
    gui_top_area_draw_buttons(widget, cr);
    gui_top_area_draw_nick_name(widget, cr, priv->nick);
    gui_top_area_draw_ip_address(widget, cr, priv->addr);
    
    cairo_destroy(cr);
	return TRUE;
}

static gboolean
gui_top_area_button_press (GtkWidget *top_area, GdkEventButton *event)
{
	GuiTopAreaPrivate *priv;

	priv = GUI_TOP_AREA_GET_PRIVATE (top_area);

	if(event->button == 1){
		if( (event->x > TOP_BUTTON_LEFT_OFFSET) && (event->x < TOP_BUTTON_SIZE+TOP_BUTTON_LEFT_OFFSET) &&
			(event->y > TOP_BUTTON_LEFT_OFFSET) && (event->y < TOP_BUTTON_SIZE+TOP_BUTTON_LEFT_OFFSET)){
			if(!priv->left_button_pushed){
				priv->left_button_pushed = TRUE;
				gtk_widget_queue_draw(top_area);
			}
			return TRUE;
		}else if( (event->x > ((top_area->allocation.width-1)-TOP_BUTTON_RIGTH_OFFSET)) &&
			(event->x < ((top_area->allocation.width-1)-TOP_BUTTON_LEFT_OFFSET)) &&
			(event->y > TOP_BUTTON_LEFT_OFFSET) &&
			(event->y < TOP_BUTTON_SIZE+TOP_BUTTON_LEFT_OFFSET)){
			if(!priv->right_button_pushed){
				priv->right_button_pushed = TRUE;
				gtk_widget_queue_draw(top_area);
			}
			return TRUE;
		}
	}

	return FALSE;
}

static gboolean
gui_top_area_button_release (GtkWidget *top_area, GdkEventButton *event)
{
	GuiTopAreaPrivate *priv;

	priv = GUI_TOP_AREA_GET_PRIVATE (GUI_TOP_AREA(top_area));

	if(event->button == 1){
		if( (event->x > TOP_BUTTON_LEFT_OFFSET) &&
			(event->x < TOP_BUTTON_SIZE+TOP_BUTTON_LEFT_OFFSET) &&
			(event->y > TOP_BUTTON_LEFT_OFFSET) &&
			(event->y < TOP_BUTTON_SIZE+TOP_BUTTON_LEFT_OFFSET)){
			g_signal_emit (GUI_TOP_AREA(top_area),
							gui_top_area_signals[SIGNAL_LEFT_BUTTON], 0);
		}else if( (event->x > ((top_area->allocation.width-1)-TOP_BUTTON_RIGTH_OFFSET)) &&
			(event->x < ((top_area->allocation.width-1)-TOP_BUTTON_LEFT_OFFSET)) &&
			(event->y > TOP_BUTTON_LEFT_OFFSET) &&
			(event->y < TOP_BUTTON_SIZE+TOP_BUTTON_LEFT_OFFSET)){
			g_signal_emit (GUI_TOP_AREA(top_area),
							gui_top_area_signals[SIGNAL_RIGHT_BUTTON], 0);
		}

		if(priv->right_button_pushed){
			priv->right_button_pushed = FALSE;
			gtk_widget_queue_draw(top_area);
		}

		if(priv->left_button_pushed){
			priv->left_button_pushed = FALSE;
			gtk_widget_queue_draw(top_area);
		}
	}

	return FALSE;
}

static gboolean
gui_top_area_motion_notify (GtkWidget *top_area, GdkEventMotion *event)
{
	GuiTopAreaPrivate *priv;

	priv = GUI_TOP_AREA_GET_PRIVATE (top_area);

	if( (event->x > TOP_BUTTON_LEFT_OFFSET) && (event->x < TOP_BUTTON_SIZE+TOP_BUTTON_LEFT_OFFSET) &&
		(event->y > TOP_BUTTON_LEFT_OFFSET) && (event->y < TOP_BUTTON_SIZE+TOP_BUTTON_LEFT_OFFSET)){
		if(!priv->left_button_over){
			priv->left_button_over = TRUE;
			gtk_widget_queue_draw(top_area);
			return TRUE;
		}
	}else if( (event->x > ((top_area->allocation.width-1)-TOP_BUTTON_RIGTH_OFFSET)) &&
			(event->x < ((top_area->allocation.width-1)-TOP_BUTTON_LEFT_OFFSET)) &&
			(event->y > TOP_BUTTON_LEFT_OFFSET) &&
			(event->y < TOP_BUTTON_SIZE+TOP_BUTTON_LEFT_OFFSET)){
		if(!priv->right_button_over){
			priv->right_button_over = TRUE;
			gtk_widget_queue_draw(top_area);
			return TRUE;
		}
	}else if(priv->left_button_over){
		priv->left_button_over = FALSE;
		gtk_widget_queue_draw(top_area);
		return TRUE;
	}else if(priv->right_button_over){
		priv->right_button_over = FALSE;
		gtk_widget_queue_draw(top_area);
		return TRUE;
	}

	return TRUE;
}

static gboolean
gui_top_area_leave_notify_event(GtkWidget *top_area, GdkEventCrossing *event)
{
	GuiTopAreaPrivate *priv;

	priv = GUI_TOP_AREA_GET_PRIVATE (top_area);

	if(priv->right_button_over || priv->left_button_over){
		priv->right_button_over = FALSE;
		priv->left_button_over = FALSE;
		gtk_widget_queue_draw(top_area);
	}

	return TRUE;
}

static void
gui_top_area_draw_nick_name(GtkWidget* widget, cairo_t *cr, gchar* nick)
{
	cairo_text_extents_t te;
	
	if(!nick) return;
	
	cairo_save(cr);		
	cairo_set_source_rgb (cr, 1, 1, 1);
	cairo_select_font_face(cr, "Georgia", CAIRO_FONT_SLANT_NORMAL, CAIRO_FONT_WEIGHT_BOLD);
	
	cairo_set_font_size(cr, 14);	
	cairo_text_extents(cr, nick, &te);
	cairo_move_to(cr, (widget->allocation.width - te.width)/2, TOP_NICK_TEXT_MARGIN + te.height);
	cairo_show_text(cr, nick);
	cairo_restore(cr);
}

static void
gui_top_area_draw_ip_address(GtkWidget* widget, cairo_t *cr, gchar* addr)
{	
	gint i;
	cairo_text_extents_t te;
	
	if(!addr) return;
	
	cairo_save(cr);		
	cairo_set_source_rgb (cr, 1, 1, 1);
	cairo_select_font_face(cr, "Georgia", CAIRO_FONT_SLANT_NORMAL, CAIRO_FONT_WEIGHT_NORMAL);
	
	for(i = 21; i > 0; i--){
		cairo_set_font_size(cr, i);
		cairo_text_extents(cr, addr, &te);
		if((te.width+(TOP_IP_TEXT_MARGIN*4)) < widget->allocation.width) break;
	}
		
	cairo_move_to(cr, (widget->allocation.width - te.width)/2, TOP_IP_TEXT_MARGIN + te.height);
	cairo_show_text(cr, addr);
	cairo_restore(cr);
}

static void
gui_top_area_draw_buttons(GtkWidget* widget, cairo_t *cr)
{	
	GuiTopAreaPrivate *priv;
	gchar* right_filename;
	gchar* left_filename;

	priv = GUI_TOP_AREA_GET_PRIVATE (widget);

	if(priv->right_button_pushed){
		right_filename = g_strconcat(PACKAGE_DATA_DIR, TOP_RIGHT_BUTTON_PUSHED, NULL);
	}else if(priv->right_button_over){
		right_filename = g_strconcat(PACKAGE_DATA_DIR, TOP_RIGHT_BUTTON_OVER, NULL);
	}else{
		right_filename = g_strconcat(PACKAGE_DATA_DIR, TOP_RIGHT_BUTTON, NULL);
	}

	if(priv->left_button_pushed){
		left_filename = g_strconcat(PACKAGE_DATA_DIR, TOP_LEFT_BUTTON_PUSHED, NULL);
	}else if(priv->left_button_over){
		left_filename = g_strconcat(PACKAGE_DATA_DIR, TOP_LEFT_BUTTON_OVER, NULL);
	}else{
		left_filename = g_strconcat(PACKAGE_DATA_DIR, TOP_LEFT_BUTTON, NULL);
	}

	gui_top_area_draw_button(widget, cr, right_filename, widget->allocation.width - TOP_BUTTON_TOTAL_SIZE - 1);
	gui_top_area_draw_button(widget, cr, left_filename, TOP_BUTTON_MARGIN);

	g_free(right_filename);
	g_free(left_filename);
}

static void
gui_top_area_draw_button(GtkWidget* widget, cairo_t *cr, gchar* path, gint x)
{	
	GdkPixbuf* img;

	img = gdk_pixbuf_new_from_file (path, NULL);

	g_return_if_fail(img);
	
	cairo_save(cr);
	gdk_cairo_set_source_pixbuf(cr, img, x, TOP_BUTTON_MARGIN);	
	cairo_paint(cr);	
	cairo_restore(cr);  
	g_object_unref (img);	
}

static void
gui_top_area_draw_base(GtkWidget *widget, cairo_t *cr)
{
	cairo_save(cr);
	cairo_set_line_width(cr, 1);
	
	cairo_set_source_rgba(cr, 0.37, 0.37, 0.37, 1);	
  	cairo_rectangle(cr, 0.5, 0.5, widget->allocation.width, widget->allocation.height);
	cairo_fill(cr);

    /** left frame extern SW -> NW -> NE */
    cairo_set_source_rgba(cr, 1, 1, 1, 1);	
  	cairo_move_to(cr, 0.5, widget->allocation.height);
  	cairo_line_to(cr, 0.5, 0.5);
  	cairo_line_to(cr, widget->allocation.width, 0.5); 	
 	cairo_stroke(cr);
 
 	/** left frame intern NE -> NW -> SW */
 	cairo_set_source_rgba(cr, 0.56, 0.56, 0.56, 1);	
	cairo_move_to(cr, 1.5, widget->allocation.height);
  	cairo_line_to(cr, 1.5, 1.5);
  	cairo_line_to(cr, widget->allocation.width, 1.5); 	
 	cairo_stroke(cr);
  
  	/** left black line NE -> SE */
  	cairo_set_source_rgba(cr, 0.0, 0.0, 0.0, 1);	
  	cairo_move_to(cr, widget->allocation.width-0.5, 0);
  	cairo_line_to(cr, widget->allocation.width-0.5, widget->allocation.height);
  	cairo_stroke(cr);

	/** right frame NE -> SE -> SW */
	cairo_set_source_rgba(cr, 0.184, 0.184, 0.184, 1);
	cairo_move_to(cr, widget->allocation.width-1.5, 1);
	cairo_line_to(cr, widget->allocation.width-1.5, widget->allocation.height-0.5);	
	cairo_line_to(cr, 1, widget->allocation.height-0.5);
	cairo_stroke(cr);
  
	cairo_restore(cr);
}

void
gui_top_area_set_nick(GuiTopArea *top_area, gchar *nick)
{
	GuiTopAreaPrivate *priv;

	priv = GUI_TOP_AREA_GET_PRIVATE (top_area);
	
	if(nick == NULL){
		if(priv->nick){ 
			g_free(priv->nick);
			priv->nick = NULL;
		}
		return;
	}
	
	if(priv->nick) g_free(priv->nick);		
	priv->nick = g_strdup(nick);
	
	/** TODO: Use gdk_window_invalidate_rect */
	gtk_widget_queue_draw(GTK_WIDGET(top_area));
}

void
gui_top_area_set_addr(GuiTopArea *top_area, gchar *addr)
{
	GuiTopAreaPrivate *priv;

	priv = GUI_TOP_AREA_GET_PRIVATE (top_area);

	if(addr == NULL){
		if(priv->addr){ 
			g_free(priv->addr);
			priv->addr = NULL;
		}
		return;
	}
	
	if(priv->addr) g_free(priv->addr);		
	priv->addr = g_strdup(addr);
	
	/** TODO: Use gdk_window_invalidate_rect */
	gtk_widget_queue_draw(GTK_WIDGET(top_area));
}

GtkWidget *
gui_top_area_new (void)
{
	return g_object_new (GUI_TYPE_TOP_AREA, NULL);
}



