/*
 *  Copyright (C) 2007 Viktor Nordell <viktor.nordell@gmail.com>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Author : Viktor Nordell <viktor.nordell@gmail.com>
 */
#include <glib.h>

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdio.h> /** perror() */

#include "io_ipc.h"

#define IPC_LIST_NETWORK_START -121
#define IPC_LIST_CLIENT_START -120

static void 		ipc_list_request		(IPCConnection* connection);
static gint			ipc_list_read_all		(IPCConnection* connection, gchar* data, gint size);
static IPCNetwork* 	ipc_list_parse			(IPCConnection* connection, gchar* data, gint size);
static IPCNetwork*	ipc_list_parse_network	(IPCConnection* connection, gchar* data);
static IPCClient* 	ipc_list_parse_client	(IPCConnection* connection, gchar* data, gint *offset);

static void
ipc_list_request(IPCConnection* connection)
{
	gchar op;
	gint size;
	
	op = IPC_GET_LIST;
	size = htonl(1);
	
	if(send(connection->socket, &size, sizeof(gint), 0) == -1){
		connection->error = IPC_ERROR_SOCKET;
		perror("send()");
		return;
	}

	if(send(connection->socket, &op, sizeof(gchar), 0) == -1){
		connection->error = IPC_ERROR_SOCKET;
		perror("send()");
		return;
	}
	
	connection->error = IPC_ERROR_NO_ERROR;
}

static IPCNetwork*
ipc_list_parse_network(IPCConnection* connection, gchar* data)
{
	guchar size;
	IPCNetwork *net;

	net = g_malloc0(sizeof(IPCNetwork));

	if(net == NULL){
		connection->error = IPC_ERROR_OTHER;
		return NULL;
	}
	
	size = data[0];
	net->name = g_strndup((data + 1), size);
	net->next = NULL;
	net->clients = NULL;		
	if(data[size+1])
		net->online = TRUE;
	else
		net->online = FALSE;
	
	connection->error = IPC_ERROR_NO_ERROR;
	
	return net;
}

static IPCClient* 
ipc_list_parse_client(IPCConnection* connection, gchar *data, gint *offset)
{	
	gchar size;
	gchar *org;
	gchar *nick;	
	guint16 port;
	gboolean online;
	IPCClient* client;
	struct in_addr *addr;
	struct in_addr *addr_tunnel = NULL;

	org = data;
	addr = (struct in_addr *)data;
	data += sizeof(struct in_addr);
	
	size = data[0];
	nick = g_strndup((data + 1), size);
	data += size + 1;
		
	if(data[0]){		
		online = TRUE;
	}else{			
		online = FALSE;
	}
	
	if(data[1]){		
		data += 2;
		addr_tunnel = (struct in_addr *)data;
		data += sizeof(struct in_addr);		
		port = ntohs((guint16)*data);
		data += sizeof(guint16);
	}else{		
		data += 2;
		port = -1;
	}
				
	client = g_malloc0(sizeof(IPCClient));
	client->nick = nick;
	client->online = online;
	client->addr = g_strdup(inet_ntoa(*addr));
	client->next = NULL;
	
	if(addr_tunnel){
		client->addr_tunnel = g_strdup(inet_ntoa(*addr_tunnel)); 
		client->port_tunnel = port;
	}
	
	connection->error = IPC_ERROR_NO_ERROR;	
	
	*offset = data - org;
	
	return client;
}

static IPCNetwork* 
ipc_list_parse(IPCConnection* connection, gchar* data, gint size)
{
	gint i;
	gint offset;
	IPCNetwork* snet = NULL;
	IPCNetwork* onet = NULL;
	IPCNetwork* cnet = NULL;
	IPCClient* oclient = NULL;
	IPCClient* cclient = NULL;	
	for(i = 0; i < size; i++){
		switch(data[i]){
			case IPC_LIST_NETWORK_START:	
				if((data[i + 1] == IPC_LIST_NETWORK_START) ||
				   (data[i + 1] == IPC_LIST_CLIENT_START)){
					i++;
					continue;
				}	
				onet = cnet;
				cnet = ipc_list_parse_network(connection, &data[++i]);
				if(onet) 
					onet->next = cnet;
				if(!snet)
					snet = cnet;
				oclient = cclient = NULL;				
			break;
			case IPC_LIST_CLIENT_START:		
				if((data[i + 1] == IPC_LIST_NETWORK_START) ||
				   (data[i + 1] == IPC_LIST_CLIENT_START)){
					i++;
					continue;
				}		
				oclient = cclient;				
				cclient = ipc_list_parse_client(connection, &data[i + 1], &offset);								
				cclient->net = cnet;
				if(oclient)
					oclient->next = cclient;
				else if(cnet)
					cnet->clients = cclient;
				i += offset;
				
			break;
			default:
			break;
		}
	}	
	return snet;
}

static gint
ipc_list_read_all(IPCConnection* connection, gchar* data, gint size)
{
	gint siz = 0;
	gint err = 0;
	gint tot = 0;

	while(TRUE){
		if((err = recv(connection->socket, &siz, sizeof(gint), 0)) == -1){
			connection->error = IPC_ERROR_SOCKET;
			perror("ipc_list_read_all - recv()");
			return -1;
		}

		siz = ntohl(siz);

		if((tot+siz) > size){
			connection->error = IPC_ERROR_ARG;
			return -1;
		}

		if((err = recv(connection->socket, data, siz, 0)) == -1){
			connection->error = IPC_ERROR_SOCKET;
			perror("ipc_list_read_all - recv()");
			return -1;
		}

		if(err != siz){
			connection->error = IPC_ERROR_SOCKET;
			return -1;
		}

		tot += siz;
		data += siz;

		if(data[-1] == -119){
			connection->error = IPC_ERROR_NO_ERROR;
			return tot;
		}
	}
}

IPCNetwork*	
ipc_list_get(IPCConnection* connection)
{
	gint err;
	gchar buffer[4096];
	IPCNetwork*	network;

	if(pthread_mutex_lock(&connection->lock) != 0)
		g_printerr("Error locking mutex\n");

	ipc_list_request(connection);
	
	if(connection->error != IPC_ERROR_NO_ERROR) 
		return NULL;		

	if((err = ipc_list_read_all(connection, (char*)&buffer, 4096)) == -1)
		return NULL;

	network = ipc_list_parse(connection, (gchar*)&buffer, err);
	
	if(pthread_mutex_unlock(&connection->lock) != 0)
		g_printerr("pthread_mutex_unlock error\n");
	
	return network;
}

void 
ipc_list_nicks_get(IPCConnection* connection)
{
	guint size = 1;
	gchar op  = IPC_GET_NICK;

	if(pthread_mutex_lock(&connection->lock) != 0)
		g_printerr("Error locking mutex\n");

	size = htonl(size);

	/** Size */
	if(send(connection->socket, &size, sizeof(guint), 0) == -1){
		connection->error = IPC_ERROR_SOCKET;
		perror("send()");
		return;
	}

	if(send(connection->socket, &op, 1, 0) == -1){
		connection->error = IPC_ERROR_SOCKET;
		perror("send()");
		return;
	}
	
	connection->error = IPC_ERROR_NO_ERROR;
	
	if(pthread_mutex_unlock(&connection->lock) != 0)
		g_printerr("pthread_mutex_unlock error\n");
}

gboolean
ipc_list_nicks_check(IPCNetwork* net)
{
	IPCClient *client;
	
	for(;net; net = net->next){		
		for(client = net->clients; client; client = client->next){
			if(!client->nick) return FALSE;
			if(strlen(client->nick) == 0) return FALSE;
		}
	}
	
	return TRUE;
}

void
ipc_list_print(IPCNetwork* net)
{
	IPCClient *client;
	
	for(;net; net = net->next){
		g_printerr(" %c [%s]\n", (net->online)?'*':' ', net->name);
		
		for(client = net->clients; client; client = client->next){
			if(client->addr_tunnel){
				g_printerr("\t %c %-13s \t%-10s\t%s:%d\n", 									
									(client->online)?'*':' ',
									client->addr,
									(client->nick)?client->nick:"",
									client->addr_tunnel,
									client->port_tunnel);
			}else{
				g_printerr("\t   %-13s \t%-10s\n", 									
									client->addr,
									(client->nick)?client->nick:"");				
			}
		}
	}	
}

void
ipc_list_free(IPCNetwork* net)
{
	IPCNetwork *old = NULL;
	IPCClient *cclient = NULL;
	IPCClient *oclient = NULL;
	
	for(;net;net = net->next){		
		if(old) g_free(old);
		if(net->name) g_free(net->name);
		
		oclient = NULL;
		for(cclient = net->clients; cclient; cclient = cclient->next){
			if(oclient) g_free(oclient);
			if(cclient->nick) g_free(cclient->nick);
			if(cclient->addr) g_free(cclient->addr);
			oclient = cclient;
		}		
		if(oclient) g_free(oclient);		
		
		old = net;		
	}
	if(old) g_free(old);
}

gboolean
ipc_list_equal(IPCNetwork* net1, IPCNetwork* net2)
{
	IPCClient *client1;
	IPCClient *client2;
	
	while(TRUE){
		if(!net1 && !net2) return TRUE;
		if(!net1) return FALSE;
		if(!net2) return FALSE;
		
		if(strcmp(net1->name, net2->name) != 0)
			return FALSE;
		if(net1->online != net2->online)
			return FALSE;
		
		client1 = net1->clients;
		client2 = net2->clients;
		
		while(TRUE){
			if(!client1 && !client2) break;
			if(!client1) return FALSE;
			if(!client2) return FALSE;
			
			if(strcmp(client1->nick, client2->nick) != 0)	
				return FALSE;
			if(client1->online != client2->online)
				return FALSE;
					
			client1 = client1->next;
			client2 = client2->next;
		}
		
		net1 = net1->next;
		net2 = net2->next;
	}	
	
	return TRUE;
}


/**
 	gint i;
	for(i = 0; i < err; i++){
		g_printerr("%4d ", buffer[i]);
	}
	g_printerr("\n");

	for(i = 0; i < err; i++){
		if((((gint) buffer[i]) >= 32) && (((gint) buffer[i]) < 127))
			g_printerr("%4c ", buffer[i]);
		else
			g_printerr("%4c ", '~');
	}
	g_printerr("\n");
 */
